# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.


class Ephemeral:
    """
    Ephemeral Messaging Layer
    """

    def __init__(self):
        pass
